"""Tests for the command-line argument parsing."""
from __future__ import absolute_import

import collections
import inspect
import os
import os.path

import losser.cli as cli

import mock

import nose.tools


# We use this in various tests to patch sys.stdout so that losser's command
# line interface doesn't errors and help text to stdout when running the tests.
DEVNULL = open(os.devnull, 'w')


# In the future we'll support simply converting all JSON fields to CSV
# columns if no columns are specified, and this test will be removed.
def test_no_columns_argument():
    """It should crash if there are no --column or --columns args."""
    table_function = mock.Mock()
    nose.tools.assert_raises(
        cli.NoColumnsError, cli.do, args=[],
        table_function=table_function)
    assert not table_function.called


def test_help():
    """It should exit with code 0 if there's a -h argument."""
    table_function = mock.Mock()
    try:
        cli.do(args=['-h'], table_function=table_function)
        assert False, "losser -h should raise an exception"
    except cli.CommandLineExit as err:
        assert err.code == 0
    assert not table_function.called


def test_long_help():
    """It should exit with code 0 if there's a --help argument."""
    table_function = mock.Mock()
    try:
        cli.do(args=['--help'], table_function=table_function)
        assert False, "losser --help should raise an exception"
    except cli.CommandLineExit as err:
        assert err.code == 0
    assert not table_function.called


def test_help_and_other_args():
    """A -h argument should override other arguments.

    It should exit with status 0 if there's a -h argument, even if there are
    other args as well.

    """
    table_function = mock.Mock()
    try:
        cli.do(
            args=['-h', '--columns', 'test_columns.json'],
            table_function=table_function)
        assert False, "losser -h should raise an exception"
    except cli.CommandLineExit as err:
        assert err.code == 0
    assert not table_function.called


def test_columns():
    """stdin, --columns and csv=True should be passed to table()."""
    table_function = mock.Mock()

    # Mock stdin with a file-like object with some JSON text input.
    mock_stdin = mock.Mock()
    mock_stdin.read.return_value = '"foobar"'

    cli.do(
        args=['--columns', 'test_columns.json'], table_function=table_function,
        in_=mock_stdin)

    table_function.assert_called_once_with(
        "foobar", "test_columns.json", csv=True, pretty=False)


@mock.patch('sys.stderr', DEVNULL)
def test_columns_with_no_arg():
    """It should crash if given a ``--columns`` option with no argument."""
    table_function = mock.Mock()

    mock_stdin = mock.Mock()
    mock_stdin.read.return_value = '"foobar"'

    try:
        cli.do(
            args=['--columns'], table_function=table_function, in_=mock_stdin)
        assert False, "It should raise if given --columns with no arg"
    except cli.CommandLineExit as err:
        assert err.code == 2

    assert not table_function.called


@mock.patch('sys.stderr', DEVNULL)
def test_unrecognized_argument():
    """It should exit with status 2 if given an unrecognized argument."""
    table_function = mock.Mock()
    try:
        cli.do(args=['--columns', 'test_columns.json', '--foobar'],
                  table_function=table_function)
        assert False, "It should raise if given an unrecognized argument"
    except cli.CommandLineExit as err:
        assert err.code == 2
    assert not table_function.called


def _absolute_path(relative_path):
    return os.path.join(os.path.dirname(os.path.abspath(
        inspect.getfile(inspect.currentframe()))), relative_path)


def test_input():
    """If given the -i argument it should read from the file not stdin."""
    mock_stdin = mock.Mock()
    mock_stdin.read.return_value = '"foobar"'

    for arg in ("-i", "--input"):
        table_function = mock.Mock()
        cli.do(
            args=['--columns', 'test_columns.json',
                  arg, _absolute_path('test_input.json')],
            table_function=table_function, in_=mock_stdin)

        assert not mock_stdin.called
        table_function.assert_called_once_with(
            "foobar", "test_columns.json", csv=True, pretty=False)


def test_with_one_column_argument():
    """Simple test with one --column and one --pattern argument."""
    table_function = mock.Mock()

    mock_stdin = mock.Mock()
    mock_stdin.read.return_value = '"foobar"'

    cli.do(
        args=["--column", "Title", "--pattern", "^title$"],
        table_function=table_function,
        in_=mock_stdin)

    table_function.assert_called_once_with(
        u"foobar",
        collections.OrderedDict(Title={"pattern": "^title$"}),
        csv=True, pretty=False)


def test_with_many_column_arguments():
    """Complex test with multiple columns and options on command line."""
    table_function = mock.Mock()

    mock_stdin = mock.Mock()
    mock_stdin.read.return_value = '"foobar"'

    cli.do(
        args=[
            "--column", "Title", "--pattern", "^title$", "--case-sensitive",
            "--strip", "no",
            "--column", "Description", "--pattern", "^notes$", "--unique",
            "--column", "Owner", "--pattern", "^author$", "--max-length=255",
            "--deduplicate", "True",
        ],
        table_function=table_function, in_=mock_stdin)

    expected_columns = collections.OrderedDict()
    expected_columns["Title"] = {
        "pattern": "^title$", "case_sensitive": True, "strip": False}
    expected_columns["Description"] = {
        "pattern": "^notes$", "unique": True}
    expected_columns["Owner"] = {
        "pattern": "^author$", "max_length": 255, "deduplicate": True}
    table_function.assert_called_once_with(
        "foobar", expected_columns, csv=True, pretty=False)


def test_with_repeated_column_option():
    """It should crash if the same option is repeated for the same column."""
    table_function = mock.Mock()
    args = ["--column", "Title", "--pattern", "^title$", "--pattern",
            "repeated"]
    nose.tools.assert_raises(
        cli.DuplicateColumnOptionError, cli.do, args=args,
        table_function=table_function)
    assert not table_function.called


def test_with_implicit_true():
    """Passing a bool like ``--unique`` with no arg should turn the option on.

    Boolean column options like unique can be turned on with ``--unique true``
    or with just ``--unique`` with no arg. This test tests the no arg version.

    """
    mock_stdin = mock.Mock()
    mock_stdin.read.return_value = '"foobar"'

    expected_columns = {
        "Title": {
            "pattern": "^title$",
            "unique": True,
        },
    }

    table_function = mock.Mock()
    cli.do(
        args=["--column", "Title", "--pattern", "^title$", "--unique"],
        table_function=table_function, in_=mock_stdin)

    table_function.assert_called_once_with(
        "foobar", expected_columns, csv=True, pretty=False)


def test_with_explicit_true():
    """Passing --case-sensitive true should turn on case-sensitive."""
    mock_stdin = mock.Mock()
    mock_stdin.read.return_value = '"foobar"'

    expected_columns = {
        "Title": {
            "pattern": "^title$",
            "case_sensitive": True,
        },
    }

    for value in ("true", "True", "y", "Y", "yes", "Yes", "YES"):
        table_function = mock.Mock()
        cli.do(
            args=[
                "--column", "Title", "--pattern", "^title$",
                "--case-sensitive", value,
            ],
            table_function=table_function,
            in_=mock_stdin
        )

        table_function.assert_called_once_with(
            "foobar", expected_columns, csv=True, pretty=False)


def test_with_explicit_false():
    """Passing --strip false should turn off strip."""
    mock_stdin = mock.Mock()
    mock_stdin.read.return_value = '"foobar"'

    expected_columns = {
        "Title": {
            "pattern": "^title$",
            "strip": False,
        },
    }

    for value in ("false", "False", "n", "N", "no", "No", "NO"):
        table_function = mock.Mock()
        cli.do(
            args=[
                "--column", "Title", "--pattern", "^title$",
                "--strip", value,
            ],
            table_function=table_function,
            in_=mock_stdin
        )

        table_function.assert_called_once_with(
            "foobar", expected_columns, csv=True, pretty=False)


def test_column_with_no_pattern():
    """It should crash if given a --column with no --pattern."""
    table_function = mock.Mock()

    mock_stdin = mock.Mock()
    mock_stdin.read.return_value = '"foobar"'

    nose.tools.assert_raises(
        cli.ColumnWithoutPatternError, cli.do,
        args=["--column", "Title"], table_function=table_function,
        in_=mock_stdin)
    assert not table_function.called


def test_column_option_with_no_column():
    """It should crash if given a column option without a preceding --column.

    """
    table_function = mock.Mock()

    mock_stdin = mock.Mock()
    mock_stdin.read.return_value = '"foobar"'

    for option in ("--unique", "--strip", "--deduplicate", "--case-sensitive",
                   "--pattern", "--max-length"):
        nose.tools.assert_raises(
            cli.ColumnOptionWithNoPrecedingColumnError,
            cli.do, args=[option, "foo"], table_function=table_function,
            in_=mock_stdin)
        assert not table_function.called


def test_boolean_option_with_invalid_arg():
    """It should crash if given a column option with an invalid argument."""
    table_function = mock.Mock()

    mock_stdin = mock.Mock()
    mock_stdin.read.return_value = '"foobar"'

    for option in ("--unique", "--strip", "--deduplicate", "--case-sensitive"):
        nose.tools.assert_raises(
            cli.InvalidColumnOptionArgument, cli.do,
            args=["--column", "foo", option, "invalid"],
            table_function=table_function, in_=mock_stdin)
        assert not table_function.called


def test_max_length():
    """Simple test of the --max-length option."""
    table_function = mock.Mock()

    mock_stdin = mock.Mock()
    mock_stdin.read.return_value = '"foobar"'

    cli.do(
        args=["--column", "Title", "--pattern", "^title$",
              "--max-length", "255"],
        table_function=table_function,
        in_=mock_stdin)

    table_function.assert_called_once_with(
        u"foobar",
        collections.OrderedDict(
            Title={"pattern": "^title$", "max_length": 255}),
        csv=True, pretty=False)


@mock.patch('sys.stderr', DEVNULL)
def test_max_length_with_no_arg():
    """It should raise if given ``--max-length`` with no argument."""
    table_function = mock.Mock()

    mock_stdin = mock.Mock()
    mock_stdin.read.return_value = '"foobar"'

    try:
        cli.do(
            args=["--column", "foo", "--pattern", "foo", "--max-length"],
            table_function=table_function, in_=mock_stdin)
        assert False, "It should raise if given --max-length with no arg"
    except cli.CommandLineExit as err:
        assert err.code == 2

    assert not table_function.called


def test_max_length_with_invalid_arg():
    """It should raise if given ``--max-length`` with an invalid argument."""
    table_function = mock.Mock()

    mock_stdin = mock.Mock()
    mock_stdin.read.return_value = '"foobar"'

    args = ["--column", "foo", "--pattern", "foo", "--max-length", "invalid"]
    nose.tools.assert_raises(
        cli.InvalidColumnOptionArgument, cli.do, args=args,
        table_function=table_function, in_=mock_stdin)

    assert not table_function.called


def test_pattern_with_multiple_arguments():
    """The --pattern option can take more than one argument."""
    table_function = mock.Mock()

    mock_stdin = mock.Mock()
    mock_stdin.read.return_value = '"foobar"'

    cli.do(
        args=["--column", "Formats", "--pattern", "^resources$", "^format$"],
        table_function=table_function,
        in_=mock_stdin)

    table_function.assert_called_once_with(
        u"foobar",
        collections.OrderedDict(
            Formats={"pattern": ["^resources$", "^format$"]}),
        csv=True, pretty=False)


def test_column_and_columns_together():
    """It should raise if given ``--column`` and ``--columns`` together."""
    table_function = mock.Mock()

    mock_stdin = mock.Mock()
    mock_stdin.read.return_value = '"foobar"'

    args = ["--column", "foo", "--pattern", "foo", "--columns", "columns.json"]
    nose.tools.assert_raises(
        cli.ColumnsAndColumnsFileError, cli.do, args=args,
        table_function=table_function, in_=mock_stdin)

    assert not table_function.called



import cStringIO
import collections
import json
import pprint
import re

import tabulate
import unicodecsv


class UniqueError(Exception):
    pass


class InvalidColumnsFileError(Exception):

    """Exception raised when reading a columns.json file fails."""

    pass


def _read_columns_file(f):
    """Return the list of column queries read from the given JSON file.

    :param f: path to the file to read
    :type f: string

    :rtype: list of dicts

    """
    try:
        columns = json.loads(open(f, 'r').read(),
                             object_pairs_hook=collections.OrderedDict)
    except Exception as err:
        raise InvalidColumnsFileError(
            "There was an error while reading {0}: {1}".format(f, err))

    # Options are not supported yet:
    if '__options' in columns:
        del columns['__options']

    return columns


def _write_csv(f, table_):
    """Write the given table (list of dicts) to the given file as CSV.

    Writes UTF8-encoded, CSV-formatted text.

    ``f`` could be an opened file, sys.stdout, or a StringIO.

    """
    # We assume that each dict in the list has the same keys.
    fieldnames = table_[0].keys()
    writer = unicodecsv.DictWriter(f, fieldnames, encoding='utf-8')
    writer.writeheader()

    # Change lists into comma-separated strings.
    for dict_ in table_:
        for key, value in dict_.items():
            if type(value) in (list, tuple):
                dict_[key] = ', '.join([unicode(v) for v in value])

    writer.writerows(table_)


def _table_to_csv(table_):
    """Return the given table converted to a CSV string.

    :param table: the table to convert
    :type table: list of OrderedDicts each with the same keys in the same
        order

    :rtype: UTF8-encoded, CSV-formatted string

    """
    f = cStringIO.StringIO()
    try:
        _write_csv(f, table_)
        return f.getvalue()
    finally:
        f.close()


def table(dicts, columns, csv=False, pretty=False):
    """Query a list of dicts with a list of queries and return a table.

    A "table" is a list of OrderedDicts each having the same keys in the same
    order.

    :param dicts: the list of input dicts
    :type dicts: list of dicts

    :param columns: the list of column query dicts, or the path to a JSON file
        containing the list of column query dicts
    :type columns: list of dicts, or string

    :param csv: return a UTF8-encoded, CSV-formatted string instead of a list
        of dicts
    :type csv: bool

    :rtype: list of dicts, or CSV string

    """
    # Optionally read columns from file.
    if isinstance(columns, basestring):
        columns = _read_columns_file(columns)

    # Either "pattern" or "pattern_path" (but not both) is allowed in the
    # columns.json file, but "pattern" gets normalised to "pattern_path" here.
    for column in columns.values():
        if "pattern" in column:
            assert "pattern_path" not in column, (
                'A column must have either a "pattern" or a "pattern_path"'
                "but not both")
            column["pattern_path"] = column["pattern"]
            del column["pattern"]


    table_ = []
    for d in dicts:
        row = collections.OrderedDict()  # The row we'll return in the table.
        for column_title, column_spec in columns.items():
            row[column_title] = query(dict_=d, **column_spec)
        table_.append(row)

    if pretty:
        # Return a pretty-printed string (looks like a nice table when printed
        # to stdout).
        return tabulate.tabulate(table_, tablefmt="grid", headers="keys")
    elif csv:
        # Return a string of CSV-formatted text.
        return _table_to_csv(table_)
    else:
        # Return a list of dicts.
        return table_


def query(pattern_path, dict_, max_length=None, strip=False,
          case_sensitive=False, unique=False, deduplicate=False,
          string_transformations=None, hyperlink=False):
    """Query the given dict with the given pattern path and return the result.

    The ``pattern_path`` is a either a single regular expression string or a
    list of regex strings that will be matched against the keys of the dict and
    its subdicts to find the value(s) in the dict to return.

    The returned result is either a single value (None, "foo", 42, False...)
    or (if the pattern path matched multiple values in the dict) a list of
    values.

    If the dict contains sub-lists or sub-dicts values from these will be
    flattened into a simple flat list to be returned.

    """
    if string_transformations is None:
        string_transformations = []

    if max_length:
        string_transformations.append(lambda x: x[:max_length])

    if hyperlink:
        string_transformations.append(
            lambda x: '=HYPERLINK("{0}")'.format(x))

    if isinstance(pattern_path, basestring):
        pattern_path = [pattern_path]

    # Copy the pattern_path because we're going to modify it which can be
    # unexpected and confusing to user code.
    original_pattern_path = pattern_path
    pattern_path = pattern_path[:]

    # We're going to be popping strings off the end of the pattern path
    # (because Python lists don't come with a convenient pop-from-front method)
    # so we need the list in reverse order.
    pattern_path.reverse()

    result = _process_object(pattern_path, dict_,
                             string_transformations=string_transformations,
                             strip=strip, case_sensitive=case_sensitive)

    if not result:
        return None  # Empty lists finally get turned into None.
    elif len(result) == 1:
        return result[0]  # One-item lists just get turned into the item.
    else:
        if unique:
            msg = "pattern_path: {0}\n\n".format(original_pattern_path)
            msg = msg + pprint.pformat(dict_)
            raise UniqueError(msg)
        if deduplicate:
            # Deduplicate the list while maintaining order.
            new_result = []
            for item in result:
                if item not in new_result:
                    new_result.append(item)
            result = new_result
        return result


def _process_object(pattern_path, object_, **kwargs):
    if type(object_) in (tuple, list):
        return _process_list(pattern_path, object_, **kwargs)
    elif isinstance(object_, dict):
        return _process_dict(pattern_path, object_, **kwargs)
    elif isinstance(object_, basestring):
        return _process_string(object_, **kwargs)
    else:
        return [object_]


def _process_string(s, string_transformations=None, strip=False, **kwargs):
    if strip:
        s = s.strip()
    if string_transformations:
        for string_transformation in string_transformations:
            s = string_transformation(s)
    return [s]


def _process_list(pattern_path, list_, **kwargs):
    result = []
    for item in list_:
        result.extend(_process_object(pattern_path[0:], item, **kwargs))
    if pattern_path:
        pattern_path.pop()
    return result


def _process_dict(pattern_path, dict_, case_sensitive=False, **kwargs):

    result = []
    pattern = pattern_path.pop()

    if case_sensitive:
        flags = re.UNICODE
    else:
        flags = re.UNICODE | re.IGNORECASE
    regex = re.compile(pattern, flags)

    for key in dict_:
        if regex.search(key):
            result.extend(_process_object(pattern_path, dict_[key],
                                          case_sensitive=case_sensitive,
                                          **kwargs))
    return result



"""The command-line interface for losser."""
from __future__ import absolute_import

import argparse
import collections
import json
import sys
import StringIO

import losser.losser as losser


class CommandLineError(Exception):

    """Exception that's raised if command-line parsing fails.

    All command-line parsing exceptions raised are either CommandLineError or
    a subclass of CommandLineError.

    """

    pass


class CommandLineExit(CommandLineError):

    """Raised when argparse has tried to make Python exit.

    For example if the user gives a --help, or if they give an unrecognized
    argument like --foobar. Argparse will already have printed the usage and
    error message to stderr, so all losser needs to do is sys.exit with the
    right exit code (as opposed to other CommandLineErrors where losser needs
    to print an error message and then sys.exit()).

    """

    def __init__(self, code):
        super(CommandLineExit, self).__init__(self)
        self.code = code


class DuplicateColumnTitleError(CommandLineError):
    pass


class ColumnOptionWithNoPrecedingColumnError(CommandLineError):
    pass


class DuplicateColumnOptionError(CommandLineError):
    pass


class ColumnWithoutPatternError(CommandLineError):
    pass


class NoColumnsError(CommandLineError):
    pass


class InvalidColumnOptionArgument(CommandLineError):
    pass


class ColumnsAndColumnsFileError(CommandLineError):
    pass


def _boolify(key, value, option_string):

    key = key.replace('-', '_')

    if not value:
        return key, True

    value = value.strip().lower()

    yes_values = ('true', 'yes', 'y')
    no_values = ('false', 'no', 'n')

    if value in yes_values:
        value = True
    elif value in no_values:
        value = False
    else:
        values = ", ".join(
            ['"{0}"'.format(value) for value in yes_values + no_values])
        raise InvalidColumnOptionArgument(
            "The argument for option {0} must be one of {1}".format(
                option_string, values))

    return key, value


def _int(key, value, option_string):

    key = key.replace('-', '_')
    try:
        value = int(value)
    except ValueError:
        raise InvalidColumnOptionArgument(
            "{0} is not a valid integer argument for {1}".format(
                value, option_string))

    return key, value


class ColumnsAction(argparse.Action):

    """An argparse action for parsing --column and related arguments.

    Saves each column and its options as a dict in an OrderedDict of columns.

    """

    def __call__(self, parser, namespace, value, option_string=None):
        if not hasattr(namespace, "columns"):
            namespace.columns = collections.OrderedDict()
        if option_string in ("-c", "--column"):

            if value in namespace.columns:
                raise DuplicateColumnTitleError(
                    "You can't have two columns called {0}".format(value))

            setattr(namespace, "__currently_parsing_column", value)
            namespace.columns[value] = {}
        else:
            if not hasattr(namespace, "__currently_parsing_column"):
                raise ColumnOptionWithNoPrecedingColumnError(
                    "You need a --column before a {0}".format(option_string))
            if option_string.startswith("--"):
                key = option_string[2:]
            else:
                assert option_string.starts_with("-")
                key = option_string[1:]
            column = namespace.columns[
                getattr(namespace, "__currently_parsing_column")]
            if key in column:
                raise DuplicateColumnOptionError(
                    "You can't have two {0}'s for the same column".format(
                        option_string))

            if key in ('case-sensitive', 'unique', 'deduplicate', 'strip'):
                key, value = _boolify(key, value, option_string)

            if key == 'max-length':
                key, value = _int(key, value, option_string)

            column[key] = value


def make_parser(add_help=True, exclude_args=None):
    """Return an argparse.ArgumentParser object with losser's arguments.

    Other projects can call this to get an ArgumentParser with losser's
    command line interface to use as a parent parser for their own parser.
    For example::

        parent_parser = losser.cli.make_parser(
            add_help=False, exclude_args=["-i"])
        parser = argparse.ArgumentParser(
            description="Export datasets from a CKAN site to JSON or CSV.",
            parents=[parent_parser])
        parser.add_argument(...

    :param add_help: Whether or not to add losser's help text to the parser.
        Pass add_help=False if you want to use your own help text in a child
        parser.
    :type add_help: bool

    :param exclude_args: List of losser command-line arguments to exclude, use
        this to exclude any default losser arguments that you don't want in
        your own command. For example: exclude_args=["-i", "--max-length"].
    :type exclude_args: list of strings

    """
    if exclude_args is None:
        exclude_args = []
    parser = argparse.ArgumentParser(add_help=add_help)
    parser.description = ("Filter, transform and export a list of JSON "
                          "objects on stdin to JSON or CSV on stdout")
    if "--columns" not in exclude_args:
        parser.add_argument(
            "--columns", dest="columns_file",
            help="the JSON file specifying the columns to be output",
        )
    if ("-i" not in exclude_args) and ("--input" not in exclude_args):
        parser.add_argument(
            "-i", "--input",
            help="read input from the given file instead of from stdin",
            dest='input_data',  # Because input is a Python builtin.
        )
    if ("-c" not in exclude_args) and ("--column" not in exclude_args):
        parser.add_argument("-c", "--column", action=ColumnsAction)
    if "--pattern" not in exclude_args:
        parser.add_argument("--pattern", action=ColumnsAction, nargs='+')
    if "--max-length" not in exclude_args:
        parser.add_argument("--max-length", action=ColumnsAction)
    if "--strip" not in exclude_args:
        parser.add_argument("--strip", nargs="?", action=ColumnsAction)
    if "--deduplicate" not in exclude_args:
        parser.add_argument("--deduplicate", nargs='?', action=ColumnsAction)
    if "--case-sensitive" not in exclude_args:
        parser.add_argument(
            "--case-sensitive", nargs='?', action=ColumnsAction)
    if "--unique" not in exclude_args:
        parser.add_argument("--unique", nargs="?", action=ColumnsAction)
    if ("-p" not in exclude_args) and ("--pretty" not in exclude_args):
        parser.add_argument("-p", "--pretty", action="store_true")
    return parser


def parse(parser=None, args=None):
    """Parse the command line arguments, return an argparse namespace object.

    Other projects can call this function and pass in their own ArgumentParser
    object (which should have a losser ArgumentParser from make_parser() above
    as parent) to do the argument parsing and get the result (this does some
    custom post-processing, beyond what argparse's parse_args() does). For
    example::

        parent_parser = losser.cli.make_parser(...)
        parser = argparse.ArgumentParser(parents=[parent_parser])
        parser.add_argument(...)
        try:
            parsed_args = losser.cli.parse(parser=parser)
        except losser.cli.CommandLineError as err:
            ...

    :raises CommandLineError: If something went wrong during command-line
        parsing. If the exception has a non-empty .message attribute it
        contains an error message that hasn't been printed to stdout yet,
        otherwise any error message has already been printed.

    :raises CommandLineExit: If the result of command-line parsing means that
        the command should exit without continuing, but this is not because of
        an error (for example if the user passed --help). Any help text will
        already have been written to stdout, the exit code that the process
        should exit with is in the exception's .code attribute.
        CommandLineExit is a subclass of CommandLineError above.

    """
    if not parser:
        parser = make_parser()

    try:
        parsed_args = parser.parse_args(args)
    except SystemExit as err:
        raise CommandLineExit(err.code)

    try:
        columns = parsed_args.columns
    except AttributeError:
        columns = collections.OrderedDict()
        parsed_args.columns = columns

    for title, spec in columns.items():
        if "pattern" not in spec:
            raise ColumnWithoutPatternError(
                'Column "{0}" needs a pattern'.format(title))

        # Change length-1 patterns into strings (not lists of one string).
        if len(spec["pattern"]) == 1:
            spec["pattern"] = spec["pattern"][0]

    if columns and parsed_args.columns_file:
        raise ColumnsAndColumnsFileError(
            "You can't use the --column and --columns options together (yet)")
    elif parsed_args.columns_file and not columns:
        parsed_args.columns = parsed_args.columns_file
    elif (not columns) and (not parsed_args.columns_file):
        # Crash if no columns specified.
        # In the future we'll support simply converting all JSON fields to CSV
        # columns if no columns are specified, and this will be removed.
        raise NoColumnsError(
            "You must give either a --columns or at least one -c/--column "
            "argument")
    else:
        assert columns

    return parsed_args


def do(parser=None, args=None, in_=None, table_function=None):
    """Read command-line args and stdin, return the result.

    Read the command line arguments and the input data from stdin, pass them to
    the table() function to do the filter and transform, and return the string
    of CSV- or JSON-formatted text that should be written to stdout.

    Note that although the output data is returned rather than written to
    stdout, this function may write error messages or help text to stdout
    (for example if there's an error with the command-line parsing).

    :raises CommandLineError: see parse() above for details

    """
    in_ = in_ or sys.stdin
    table_function = table_function or losser.table

    parsed_args = parse(parser=parser, args=args)

    # Read the input data from stdin or a file.
    if parsed_args.input_data:
        input_data = open(parsed_args.input_data, 'r').read()
    else:
        input_data = in_.read()
    dicts = json.loads(input_data)

    csv_string = table_function(dicts, parsed_args.columns, csv=True,
                                pretty=parsed_args.pretty)

    return csv_string


def main():
    """Call do() and if it raises an exception then sys.exit() appropriately.

    This makes sure that any usage and error messages are printed correctly,
    and that the exit code is right.

    do() itself doesn't call sys.exit() because we want it to be callable from
    tests that check that it raises the right exception classes for different
    invalid inputs.

    """
    parser = make_parser()
    try:
        output = do(parser=parser)
    except CommandLineExit as err:
        sys.exit(err.code)
    except CommandLineError as err:
        if err.message:
            parser.error(err.message)
    sys.stdout.write(output)


if __name__ == "__main__": main()
